import json
import logging
import os
import os.path
import pathlib
import shutil
import stat
import sys
import tarfile
import zipfile
from pathlib import Path
from typing import Any, Dict, List, Union

import toml

from erdpy import errors

logger = logging.getLogger("utils")


class Object:
    def __repr__(self):
        return str(self.__dict__)

    def to_json(self):
        data_json = json.dumps(self.__dict__, indent=4)
        return data_json


class ObjectEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Object):
            return obj.__dict__
        return json.JSONEncoder.default(self, obj)


def omit_fields(data: Any, fields: List[str] = []):
    if isinstance(data, dict):
        for field in fields:
            data.pop(field, None)
        return data
    raise errors.ProgrammingError("omit_fields: only dictionaries are supported.")


def untar(archive_path: str, destination_folder: str) -> None:
    logger.debug(f"untar [{archive_path}] to [{destination_folder}].")

    ensure_folder(destination_folder)
    tar = tarfile.open(archive_path)
    tar.extractall(path=destination_folder)
    tar.close()

    logger.debug("untar done.")


def unzip(archive_path, destination_folder):
    logger.debug(f"unzip [{archive_path}] to [{destination_folder}].")

    ensure_folder(destination_folder)
    with zipfile.ZipFile(archive_path, "r") as my_zip:
        my_zip.extractall(destination_folder)

    logger.debug("unzip done.")


def ensure_folder(folder: Union[str, Path]):
    pathlib.Path(folder).mkdir(parents=True, exist_ok=True)


def read_lines(file: str):
    with open(file) as f:
        lines = f.readlines()
    lines = [line.strip() for line in lines]
    lines = [line for line in lines if line]
    return lines


def read_file(f: Any, binary=False) -> Union[str, bytes]:
    try:
        mode = "rb" if binary else "r"
        if isinstance(f, str) or isinstance(f, pathlib.PosixPath):
            with open(f, mode) as f:
                return f.read()
        return f.read()
    except Exception as err:
        raise errors.BadFile(f, err)


def write_file(f: Any, text: str):
    if isinstance(f, str) or isinstance(f, pathlib.PosixPath):
        with open(f, "w") as f:
            return f.write(text)
    return f.write(text)


def read_toml_file(filename):
    return toml.load(str(filename))


def write_toml_file(filename, data):
    with open(str(filename), "w") as f:
        toml.dump(data, f)


def read_json_file(filename: str) -> Dict[str, Any]:
    with open(filename) as f:
        return json.load(f)


def write_json_file(filename: str, data: Any):
    with open(filename, "w") as f:
        json.dump(data, f, indent=4)


def dump_out_json(data: Any, outfile: Any = None):
    if not outfile:
        outfile = sys.stdout

    json.dump(data, outfile, indent=4, cls=ObjectEncoder)
    outfile.write("\n")


def get_subfolders(folder):
    return [item.name for item in os.scandir(folder) if item.is_dir() and not item.name.startswith(".")]


def mark_executable(file: str) -> None:
    logger.debug(f"Mark [{file}] as executable")
    st = os.stat(file)
    os.chmod(file, st.st_mode | stat.S_IEXEC)


def find_in_dictionary(dictionary, compound_path):
    keys = compound_path.split(".")
    node = dictionary
    for key in keys:
        node = node.get(key)
        if node is None:
            break

    return node


def list_files(folder: str, suffix: str = None) -> List[str]:
    files = os.listdir(folder)
    files = [os.path.join(folder, f) for f in files]

    if suffix:
        files = [e for e in files if e.lower().endswith(suffix.lower())]

    return files


def remove_folder(folder):
    shutil.rmtree(folder, ignore_errors=True)


def symlink(real: str, link: str) -> None:
    if os.path.exists(link):
        os.remove(link)
    os.symlink(real, link)


def str_to_bool(input: str) -> bool:
    return str(input).lower() in ["true", "1", "t", "y", "yes"]


def as_object(input: Any) -> Object:
    if isinstance(input, dict):
        result = Object()
        result.__dict__.update(input)
        return result

    return input


def is_arg_present(key: str, args: List[str]) -> bool:
    for arg in args:
        if arg.find("--data") != -1:
            continue
        if arg.find(key) != -1:
            return True

    return False


# https://code.visualstudio.com/docs/python/debugging
# https://code.visualstudio.com/docs/python/debugging
def breakpoint():
    import debugpy
    debugpy.listen(5678)
    print("Waiting for debugger attach")
    debugpy.wait_for_client()
    debugpy.breakpoint()



import logging
from typing import Any, List, Tuple

from erdpy import transactions, utils
from erdpy.accounts import Address
from erdpy.proxy.http_facade import do_get, do_post
from erdpy.proxy.messages import NetworkConfig

METACHAIN_ID = 4294967295
ANY_SHARD_ID = 0

logger = logging.getLogger("proxy")


class ElrondProxy:
    def __init__(self, url: str):
        self.url = url

    def get_account_nonce(self, address: Address) -> int:
        url = f"{self.url}/address/{address.bech32()}"
        response = do_get(url)
        nonce = response.get("account").get("nonce", 0)
        return int(nonce)

    def get_account_balance(self, address: Address):
        url = f"{self.url}/address/{address.bech32()}/balance"
        response = do_get(url)
        balance = response.get("balance", 0)
        return int(balance)

    def get_account(self, address: Address):
        url = f"{self.url}/address/{address.bech32()}"
        response = do_get(url)
        account = response.get("account", dict())
        return account

    def get_account_transactions(self, address: Address):
        TRUNCATE_DATA_THRESHOLD = 75

        url = f"{self.url}/address/{address.bech32()}/transactions"
        response = do_get(url)
        transactions = response.get("transactions", [])
        for transaction in transactions:
            data = transaction.get("data", "")
            data = (data[:TRUNCATE_DATA_THRESHOLD] + ' ... truncated ...') if len(data) > TRUNCATE_DATA_THRESHOLD else data
            transaction["data"] = data
        return transactions

    def get_num_shards(self):
        network_config = self.get_network_config()
        return network_config.num_shards

    def get_last_block_nonce(self, shard_id):
        if shard_id == "metachain":
            metrics = self._get_network_status(METACHAIN_ID)
        else:
            metrics = self._get_network_status(shard_id)

        nonce = metrics.get("erd_highest_final_nonce", 0)
        return nonce

    def get_gas_price(self):
        network_config = self.get_network_config()
        return network_config.min_gas_price

    def get_chain_id(self):
        network_config = self.get_network_config()
        return network_config.chain_id

    def _get_network_status(self, shard_id):
        url = f"{self.url}/network/status/{shard_id}"
        response = do_get(url)
        payload = response.get("status")
        return payload

    def get_network_config(self) -> NetworkConfig:
        url = f"{self.url}/network/config"
        response = do_get(url)
        payload = response.get("config")
        result = NetworkConfig(payload)
        return result

    def send_transaction(self, payload: Any) -> str:
        url = f"{self.url}/transaction/send"
        response = do_post(url, payload)
        tx_hash = response.get("txHash")
        return tx_hash

    def simulate_transaction(self, payload: Any) -> str:
        url = f"{self.url}/transaction/simulate"
        utils.dump_out_json(payload)
        response = do_post(url, payload)
        return response

    def send_transactions(self, payload: List[Any]) -> Tuple[int, List[str]]:
        url = f"{self.url}/transaction/send-multiple"
        response = do_post(url, payload)
        # Proxy and Observers have different response format:
        num_sent = response.get("numOfSentTxs", 0) or response.get("txsSent", 0)
        hashes = response.get("txsHashes")
        return num_sent, hashes

    def query_contract(self, payload: Any):
        url = f"{self.url}/vm-values/query"
        response = do_post(url, payload)
        return response

    def get_transaction(self, tx_hash: str, sender_address: str = "") -> Any:
        url = f"{self.url}/transaction/{tx_hash}"
        if sender_address:
            url = f"{self.url}/transaction/{tx_hash}?sender={sender_address}"

        response = do_get(url)
        transaction = response.get("transaction", dict())
        return transaction

    def get_hyperblock(self, key) -> Any:
        url = f"{self.url}/hyperblock/by-hash/{key}"
        if str(key).isnumeric():
            url = f"{self.url}/hyperblock/by-nonce/{key}"

        response = do_get(url)
        return response



from typing import Any

from erdpy.testnet.config import TestnetConfiguration


def patch(data: Any, testnet_config: TestnetConfiguration):
    data['DbLookupExtensions']['Enabled'] = True
    data['GeneralSettings']['SCDeployEnableEpoch'] = 0
    data['GeneralSettings']['BuiltInFunctionsEnableEpoch'] = 0
    data['GeneralSettings']['RelayedTransactionsEnableEpoch'] = 0
    data['GeneralSettings']['PenalizedTooMuchGasEnableEpoch'] = 0
    data['GeneralSettings']['SwitchJailWaitingEnableEpoch'] = 0


def patch_api(data: Any, testnet_config: TestnetConfiguration):
    routes = data['APIPackages']['transaction']['Routes']
    for route in routes:
        route["Open"] = True


def patch_system_smart_contracts(data: Any, testnet_config: TestnetConfiguration):
    data['StakingSystemSCConfig']['StakeEnableEpoch'] = 0
    data['StakingSystemSCConfig']['ActivateBLSPubKeyMessageVerification'] = True
    data['ESDTSystemSCConfig']['EnabledEpoch'] = 0




from erdpy import cli_shared
import logging
import sys
from typing import Any

from erdpy import config, utils

logger = logging.getLogger("cli.config")


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "config", "Configure elrond-sdk (default values etc.)")
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(subparsers, "config", "dump", "Dumps configuration.")
    sub.set_defaults(func=dump)

    sub = cli_shared.add_command_subparser(subparsers, "config", "get", "Gets a configuration value.")
    _add_name_arg(sub)
    sub.set_defaults(func=get_value)

    sub = cli_shared.add_command_subparser(subparsers, "config", "set", "Sets a configuration value.")
    _add_name_arg(sub)
    sub.add_argument("value", help="the new value")
    sub.set_defaults(func=set_value)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def _add_name_arg(sub: Any):
    sub.add_argument("name", help="the name of the configuration entry")


def dump(args: Any):
    data = config.read_file()
    utils.dump_out_json(data, sys.stdout)


def get_value(args: Any):
    value = config.get_value(args.name)
    print(value)


def set_value(args: Any):
    config.set_value(args.name, args.value)



import asyncio
import logging
import subprocess
import traceback
from typing import Any, List

from erdpy import errors, feedback

logger = logging.getLogger("myprocess")


def run_process(args: List[str], env: Any = None, dump_to_stdout: bool = True, cwd: str = None):
    logger.info(f"run_process: {args}, in folder: {cwd}")

    try:
        output = subprocess.check_output(args, shell=False, universal_newlines=True, stderr=subprocess.STDOUT, env=env, cwd=cwd)
        logger.info("Successful run. Output:")
        if dump_to_stdout:
            print(output or "[No output]")
        return output
    except subprocess.CalledProcessError as error:
        raise errors.ExternalProcessError(error.cmd, error.output)


def run_process_async(args: List[str], env: Any = None, cwd: str = None):
    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(async_subprocess(args, env=env, sinks=None, cwd=cwd))
    loop.close()
    asyncio.set_event_loop(asyncio.new_event_loop())
    return result


async def async_subprocess(args, env=None, sinks=None, cwd: str = None):
    process = await asyncio.create_subprocess_exec(*args, env=env, stdout=asyncio.subprocess.PIPE,
                                                   stderr=asyncio.subprocess.PIPE, cwd=cwd)

    await asyncio.wait([
        _read_stream(process.stdout, sinks),
        _read_stream(process.stderr, sinks)
    ])
    return await process.wait()


async def _read_stream(stream, sinks=None):
    while True:
        try:
            line = await stream.readline()
            if line:
                line = line.decode("utf-8", "replace").strip()
                if sinks is None:
                    print(line)
                else:
                    for sink in sinks:
                        feedback.get_sink(sink).write(line)
            else:
                break
        except Exception:
            print(traceback.format_exc())



import base64
import itertools
import textwrap
from os import path
from pathlib import Path
from typing import Tuple, Union

from erdpy import guards, utils


def get_pubkey(pem_file: Union[str, Path]):
    _, pubkey = parse(pem_file)
    return pubkey


def parse(pem_file: Union[str, Path], index: int = 0) -> Tuple[bytes, bytes]:
    pem_file = path.expanduser(pem_file)
    guards.is_file(pem_file)

    lines = utils.read_lines(pem_file)
    keys_lines = [list(key_lines) for is_next_key, key_lines in itertools.groupby(lines, lambda line: "-----" in line) if not is_next_key]
    keys = ["".join(key_lines) for key_lines in keys_lines]

    key_base64 = keys[index]
    key_hex = base64.b64decode(key_base64).decode()
    key_bytes = bytes.fromhex(key_hex)

    seed = key_bytes[:32]
    pubkey = key_bytes[32:]
    return seed, pubkey


def parse_validator_pem(pem_file):
    pem_file = path.expanduser(pem_file)
    guards.is_file(pem_file)

    lines = utils.read_lines(pem_file)

    first_line = lines[0]
    start = first_line.find('for ') + len('for ')
    stop = first_line.rfind('-----')
    bls_key = first_line[start:stop]

    lines = [line for line in lines if "-----" not in line]
    key_base64 = "".join(lines)
    key_hex = base64.b64decode(key_base64).decode()
    key_bytes = bytes.fromhex(key_hex)

    seed = key_bytes
    return seed, bls_key


def write(pem_file: Union[str, Path], seed: bytes, pubkey: bytes, name: str = ""):
    pem_file = path.expanduser(pem_file)

    if not name:
        name = pubkey.hex()

    header = f"-----BEGIN PRIVATE KEY for {name}-----"
    footer = f"-----END PRIVATE KEY for {name}-----"

    seed_hex = seed.hex()
    pubkey_hex = pubkey.hex()
    combined = seed_hex + pubkey_hex
    combined = combined.encode()
    key_base64 = base64.b64encode(combined).decode()

    payload_lines = textwrap.wrap(key_base64, 64)
    payload = "\n".join(payload_lines)
    content = "\n".join([header, payload, footer])
    utils.write_file(pem_file, content)



import logging
import nacl.encoding
import nacl.signing
import base64
from os import path

from erdpy import utils, guards

logger = logging.getLogger("wallet.generator")


def generate_pairs():
    pass


def generate_pair():
    signing_key = nacl.signing.SigningKey.generate()
    seed_bytes = bytes(signing_key)
    pubkey_bytes = bytes(signing_key.verify_key)
    return seed_bytes, pubkey_bytes



from erdpy.proxy.core import ElrondProxy
from erdpy.proxy.cost import TransactionCostEstimator
from erdpy.proxy.tx_types import TxTypes

__all__ = ["ElrondProxy", "TransactionCostEstimator", "TxTypes"]




import logging
import os
import shutil
from os import path
from typing import Dict, List

from erdpy import config, downloader, errors, myprocess, utils, workstation

logger = logging.getLogger("modules")


class DependencyModule:
    def __init__(self, key: str, aliases: List[str]):
        self.key = key
        self.aliases = aliases

    def get_directory(self, tag: str) -> str:
        raise NotImplementedError()

    def install(self, tag: str, overwrite: bool) -> None:
        # Fallback to default tag if not provided
        tag = tag or config.get_dependency_tag(self.key)

        logger.debug(f"install: key={self.key}, tag={tag}")

        if self._should_skip(tag, overwrite):
            logger.debug("Already exists. Skip install.")
            return

        self.uninstall(tag)
        self._do_install(tag)

        # Upon installation we update the default tag
        config.set_dependency_tag(self.key, tag)
        self._post_install(tag)

    def _do_install(self, tag: str) -> None:
        raise NotImplementedError()

    def _post_install(self, tag: str):
        pass

    def _should_skip(self, tag: str, overwrite: bool) -> bool:
        if overwrite:
            return False
        return self.is_installed(tag)

    def uninstall(self, tag: str) -> None:
        raise NotImplementedError()

    def is_installed(self, tag: str) -> bool:
        raise NotImplementedError()

    def get_env(self) -> Dict[str, str]:
        raise NotImplementedError()


class StandaloneModule(DependencyModule):

    def __init__(self, key: str, aliases: List[str] = None):
        if aliases is None:
            aliases = list()

        super().__init__(key, aliases)
        self.archive_type = "tar.gz"

    def _do_install(self, tag: str):
        self._download(tag)
        self._extract(tag)

    def uninstall(self, tag: str):
        if os.path.isdir(self.get_directory(tag)):
            shutil.rmtree(self.get_directory(tag))

    def is_installed(self, tag: str):
        return path.isdir(self.get_directory(tag))

    def _download(self, tag: str):
        url = self._get_download_url(tag)
        archive_path = self._get_archive_path(tag)
        downloader.download(url, archive_path)

    def _extract(self, tag: str):
        archive_path = self._get_archive_path(tag)
        destination_folder = self.get_directory(tag)

        if self.archive_type == "tar.gz":
            utils.untar(archive_path, destination_folder)
        elif self.archive_type == "zip":
            utils.unzip(archive_path, destination_folder)
        else:
            raise errors.UnknownArchiveType(self.archive_type)

    def get_directory(self, tag: str):
        folder = path.join(self.get_parent_directory(), tag)
        return folder

    def get_parent_directory(self):
        tools_folder = workstation.get_tools_folder()
        return path.join(tools_folder, self.key)

    def _get_download_url(self, tag: str) -> str:
        platform = workstation.get_platform()
        url = config.get_dependency_url(self.key, tag, platform)
        if url is None:
            raise errors.PlatformNotSupported(self.key, platform)

        url = url.replace("{TAG}", tag)
        return url

    def _get_archive_path(self, tag: str) -> str:
        tools_folder = workstation.get_tools_folder()
        archive = path.join(tools_folder, f"{self.key}.{tag}.{self.archive_type}")
        return archive


class ArwenToolsModule(StandaloneModule):
    def __init__(self, key: str, aliases: List[str] = None):
        if aliases is None:
            aliases = list()

        super().__init__(key, aliases)

    def _post_install(self, tag: str):
        directory = self.get_directory(tag)

        utils.mark_executable(path.join(directory, "arwen"))
        utils.mark_executable(path.join(directory, "arwendebug"))
        utils.mark_executable(path.join(directory, "test"))

        utils.symlink(path.join(directory, "arwendebug"), os.path.join(self.get_parent_directory(), "arwendebug"))
        utils.symlink(path.join(directory, "test"), os.path.join(self.get_parent_directory(), "mandos-test"))

    def get_env(self):
        return {
        }


class GolangModule(StandaloneModule):
    def __init__(self, key: str, aliases: List[str] = None):
        if aliases is None:
            aliases = list()

        super().__init__(key, aliases)

    def _post_install(self, tag: str):
        parent_directory = self.get_parent_directory()
        utils.ensure_folder(path.join(parent_directory, "GOPATH"))
        utils.ensure_folder(path.join(parent_directory, "GOCACHE"))

    def get_env(self):
        directory = self.get_directory(config.get_dependency_tag(self.key))
        parent_directory = self.get_parent_directory()

        return {
            "PATH": f"{path.join(directory, 'go/bin')}:{os.environ['PATH']}",
            "GOPATH": self.get_gopath(),
            "GOCACHE": path.join(parent_directory, "GOCACHE"),
            "GOROOT": path.join(directory, "go")
        }

    def get_gopath(self):
        return path.join(self.get_parent_directory(), "GOPATH")


class NodejsModule(StandaloneModule):
    def __init__(self, key: str, aliases: List[str]):
        super().__init__(key, aliases)

    def _post_install(self, tag: str):
        # We'll create a symlink towards the payload folder
        subfolder_to_bypass = self._get_download_url(tag).split("/")[-1]
        subfolder_to_bypass = subfolder_to_bypass.replace(f".{self.archive_type}", "")
        payload_folder = path.join(self.get_directory(tag), subfolder_to_bypass)
        link = path.join(self.get_parent_directory(), "latest")

        utils.symlink(payload_folder, link)

    def get_env(self):
        bin_folder = path.join(self.get_parent_directory(), "latest", "bin")

        return {
            "PATH": f"{bin_folder}:{os.environ['PATH']}",
        }


class Rust(DependencyModule):
    def __init__(self, key: str, aliases: List[str] = None):
        if aliases is None:
            aliases = list()

        super().__init__(key, aliases)

    def _do_install(self, tag: str) -> None:
        rustup_path = self._get_rustup_path()
        downloader.download("https://sh.rustup.rs", rustup_path)
        utils.mark_executable(rustup_path)

        args = [rustup_path, "--verbose", "--default-toolchain", "nightly", "--profile",
                "minimal", "--target", "wasm32-unknown-unknown", "--no-modify-path", "-y"]
        myprocess.run_process_async(args, env=self.get_env())

    def uninstall(self, tag: str):
        directory = self.get_directory("")
        if os.path.isdir(directory):
            shutil.rmtree(directory)

    def is_installed(self, tag: str):
        try:
            myprocess.run_process(["rustc", "--version"], env=self.get_env())
            return True
        except Exception:
            return False

    def _get_rustup_path(self):
        tools_folder = workstation.get_tools_folder()
        return path.join(tools_folder, "rustup.sh")

    def get_directory(self, tag: str):
        tools_folder = workstation.get_tools_folder()
        return path.join(tools_folder, "vendor-rust")

    def get_env(self):
        directory = self.get_directory("")

        return {
            "PATH": f"{path.join(directory, 'bin')}:{os.environ['PATH']}",
            "RUSTUP_HOME": directory,
            "CARGO_HOME": directory
        }


class MclSignerModule(StandaloneModule):
    def __init__(self, key: str, aliases: List[str] = None):
        if aliases is None:
            aliases = list()

        super().__init__(key, aliases)

    def _post_install(self, tag: str):
        directory = self.get_directory(tag)
        utils.mark_executable(path.join(directory, "signer"))



import asyncio
import logging
import traceback
from typing import Any

from erdpy.testnet.config import TestnetConfiguration

logger = logging.getLogger("testnet")


NODES_START_DELAY = 1
PROXY_START_DELAY = 10

is_after_genesis = False


def start(args: Any):
    try:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(do_start(args))
        loop.close()
        asyncio.set_event_loop(asyncio.new_event_loop())
    except KeyboardInterrupt:
        pass


async def do_start(args: Any):
    testnet_config = TestnetConfiguration.from_file(args.configfile)
    logger.info('testnet folder is %s', testnet_config.root())

    to_run = []

    # Seed node
    to_run.append(run(["./seednode", "--log-save"], cwd=testnet_config.seednode_folder()))

    loglevel = _patch_loglevel(testnet_config.loglevel())

    # Observers
    for observer in testnet_config.observers():
        to_run.append(run([
            "./node",
            "--use-log-view",
            "--log-save",
            f"--log-level={loglevel}",
            "--log-logger-name",
            "--log-correlation",
            f"--destination-shard-as-observer={observer.shard}",
            f"--rest-api-interface=localhost:{observer.api_port}"
        ], cwd=observer.folder, delay=NODES_START_DELAY))

    # Validators
    for validator in testnet_config.validators():
        to_run.append(run([
            "./node",
            "--use-log-view",
            "--log-save",
            f"--log-level={loglevel}",
            "--log-logger-name",
            "--log-correlation",
            f"--rest-api-interface=localhost:{validator.api_port}"
        ], cwd=validator.folder, delay=NODES_START_DELAY))

    # Proxy
    to_run.append(run([
        "./proxy",
    ], cwd=testnet_config.proxy_folder(), delay=PROXY_START_DELAY))

    await asyncio.gather(*to_run)


async def run(args, env=None, cwd: str = None, delay: int = 0):
    await asyncio.sleep(delay)

    process = await asyncio.create_subprocess_exec(*args, env=env, stdout=asyncio.subprocess.PIPE,
                                                   stderr=asyncio.subprocess.PIPE, cwd=cwd)

    pid = process.pid

    print(f"Started process [{pid}]", args)
    await asyncio.wait([
        _read_stream(process.stdout, pid),
        _read_stream(process.stderr, pid)
    ])

    return_code = await process.wait()
    print(f"Proces [{pid}] stopped. Return code: {return_code}.")


async def _read_stream(stream, pid):
    while True:
        try:
            line = await stream.readline()
            if line:
                line = line.decode("utf-8", "replace").strip()
                if _is_interesting_logline(line):
                    _dump_interesting_log_line(pid, line)
            else:
                break
        except Exception:
            print(traceback.format_exc())


def _patch_loglevel(loglevel: str) -> str:
    loglevel = loglevel or "*:DEBUG"
    if "arwen/host:" not in loglevel:
        loglevel += ",arwen/host:TRACE"
    return loglevel


def _is_interesting_logline(logline):
    global is_after_genesis

    if "started committing block" in logline:
        is_after_genesis = True

    if not is_after_genesis:
        return any(e in logline for e in ["started committing block", "ERROR", "WARN"])    
    return any(e in logline for e in ["started committing block", "ERROR", "WARN", "arwen"])


def _dump_interesting_log_line(pid: str, logline: str) -> str:
    print(f"[PID={pid}]", logline)



import json
import logging
import os
import time
from collections import OrderedDict
from os import path

from erdpy import utils
from erdpy.accounts import Account
from erdpy.proxy.core import ElrondProxy
from erdpy.transactions import BunchOfTransactions
from erdpy.workstation import get_tools_folder

logger = logging.getLogger("queue")


def _prepare_tx(args):
    ordered_fields = OrderedDict()
    ordered_fields["nonce"] = 0
    ordered_fields["value"] = int(args.value)
    ordered_fields["receiver"] = args.receiver
    ordered_fields["gasPrice"] = int(args.gas_price)
    ordered_fields["gasLimit"] = int(args.gas_limit)
    ordered_fields["data"] = args.data
    ordered_fields["chain"] = args.chain
    ordered_fields["version"] = args.version

    return ordered_fields


def _wait_to_execute_txs(proxy, owner, expected_nonce):
    new_nonce = 0
    old_time = time.time()
    timeout = False
    while new_nonce < expected_nonce and not timeout:
        owner.sync_nonce(proxy)
        new_nonce = owner.nonce
        # timeout if passed a minute
        timeout = time.time() - old_time > 59
    if timeout:
        logger.error("Not all transactions were executed")
    else:
        logger.info(f"Transactions were executed")


class TransactionQueue:
    _TXS_FILE_NAME = "txs.json"
    _TXS_FIELD_NAME = "transactions"
    _TXS_INFO_FILE_NAME = "txs_info.txt"

    def __init__(self):
        tools_folder = get_tools_folder()

        txs_file_dir = path.join(tools_folder, "transactions")
        # create transactions directory if not exits
        if not os.path.exists(txs_file_dir):
            os.mkdir(txs_file_dir)

        self.txs_file_path = path.join(txs_file_dir, self._TXS_FILE_NAME)
        if not os.path.exists(self.txs_file_path):
            # create transactions file if not exits
            utils.write_file(self.txs_file_path, '{"' + self._TXS_FIELD_NAME + '":[]}')

        self.txs_info_file_path = path.join(txs_file_dir, self._TXS_INFO_FILE_NAME)
        if not os.path.exists(self.txs_info_file_path):
            utils.write_file(self.txs_info_file_path, f"index:{0}")

    def enqueue_transaction(self, args):
        prepared = _prepare_tx(args)
        data = self._read_json_file()
        temp_data = data[self._TXS_FIELD_NAME]
        temp_data.append(prepared)
        self._write_json_file(data)

    def _read_json_file(self):
        with open(self.txs_file_path, "r") as json_file:
            try:
                data = json.load(json_file)
            except Exception:
                logger.error("Cannot read transactions file")
                return {self._TXS_FIELD_NAME: []}
            return data

    def _write_json_file(self, data):
        with open(self.txs_file_path, "w") as json_file:
            json.dump(data, json_file, indent=1)

    def _remove_all_transactions(self):
        os.remove(self.txs_file_path)

    def _read_index(self):
        info = utils.read_file(self.txs_info_file_path)
        info_slit = info.split(":")
        return int(info_slit[1])

    def clean_transactions_queue(self):
        os.remove(self.txs_info_file_path)
        os.remove(self.txs_file_path)

    def dispatch_transactions_continuously(self, args):
        while True:
            logger.info("dispatch_transactions_continuously()")
            self.dispatch_transactions(args)
            time.sleep(int(args.interval))

    def dispatch_transactions(self, args):
        data = self._read_json_file()
        txs = data[self._TXS_FIELD_NAME]
        txs_index = self._read_index()

        total_txs = len(txs) - txs_index
        if total_txs == 0 or len(txs) == 0:
            logger.info("No transactions to dispatch")
            return

        proxy = ElrondProxy(args.proxy)
        # Need to sync nonce
        if args.pem:
            owner = Account(pem_file=args.pem)
        elif args.keyfile and args.passfile:
            owner = Account(key_file=args.keyfile, pass_file=args.passfile)

        owner.sync_nonce(proxy)
        nonce = owner.nonce
        old_nonce = nonce

        print(nonce)
        bunch = BunchOfTransactions()
        idx = txs_index
        while idx < len(txs):
            tx = txs[idx]
            # TODO CHECK IF BUNCH OF TRANSACTION generate transactions with chain id and version
            bunch.add(
                owner,
                tx.get("receiver"),
                nonce,
                tx.get("value"),
                tx.get("data"),
                tx.get("gasPrice"),
                tx.get("gasLimit")
            )
            # increment nonce
            nonce += 1
            idx += 1

        logger.info(f"Sending {total_txs} transactions")
        try:
            num_sent, hashes = bunch.send(proxy)
        except Exception:
            logger.error("No valid transactions to send")
            num_sent = 0
            hashes = []

        logger.info(f"{num_sent} transactions were accepted by observers")
        for key in hashes:
            print(f"tx {txs_index+int(key)}: hash => {hashes[key]}")

        utils.write_file(self.txs_info_file_path, f"index:{len(txs)}")
        # wait until transactions are executed
        _wait_to_execute_txs(proxy, owner, old_nonce + num_sent)



import json
import logging
import os
from os import path

from erdpy import errors, utils
from erdpy.projects import shared
from erdpy.projects.project_rust import CargoFile
from erdpy.projects.templates_config import get_templates_repositories

logger = logging.getLogger("projects.templates")


def list_project_templates():
    templates = []

    for repository in get_templates_repositories():
        repository.download()
        for template in repository.get_templates():
            templates.append(TemplateSummary(template, repository))

    templates = sorted(templates, key=lambda item: item.name)

    pretty_json = json.dumps([item.__dict__ for item in templates], indent=4)
    print(pretty_json)


class TemplateSummary():
    def __init__(self, name, repository):
        self.name = name
        self.github = repository.github
        self.language = repository.get_language(name)


def create_from_template(name: str, template_name: str, directory: str):
    directory = path.expanduser(directory)

    logger.info("create_from_template.name: %s", name)
    logger.info("create_from_template.template_name: %s", template_name)
    logger.info("create_from_template.directory: %s", directory)

    if not directory:
        logger.info("Using current directory")
        directory = os.getcwd()

    project_directory = path.join(directory, name)
    if path.exists(project_directory):
        raise errors.BadDirectory(project_directory)

    _download_templates_repositories()
    _copy_template(template_name, project_directory)

    template = _load_as_template(project_directory)
    template.apply(template_name, name)

    logger.info("Project created, template applied.")


def _download_templates_repositories():
    for repo in get_templates_repositories():
        repo.download()


def _copy_template(template, destination_path):
    for repo in get_templates_repositories():
        if repo.has_template(template):
            repo.copy_template(template, destination_path)
            return

    raise errors.TemplateMissingError(template)


def _load_as_template(directory):
    if shared.is_source_clang(directory):
        return TemplateClang(directory)
    if shared.is_source_sol(directory):
        return TemplateSol(directory)
    if shared.is_source_rust(directory):
        return TemplateRust(directory)


class Template:
    def __init__(self, directory):
        self.directory = directory

    def apply(self, template_name, project_name):
        self.template_name = template_name
        self.project_name = project_name
        self._patch()

    def _patch(self):
        """Implemented by derived classes"""
        pass


class TemplateClang(Template):
    pass


class TemplateRust(Template):
    CARGO_TOML = "Cargo.toml"

    def _patch(self):
        logger.info("Patching cargo files...")
        self._patch_cargo()
        self._patch_cargo_wasm()
        self._patch_cargo_debug()

        logger.info("Patching source code...")
        self._patch_source_code_wasm()
        self._patch_source_code_debug()
        self._patch_source_code_tests()

        logger.info("Patching test files...")
        self._patch_mandos_tests()

    def _patch_cargo(self):
        cargo_path = path.join(self.directory, TemplateRust.CARGO_TOML)

        cargo_file = CargoFile(cargo_path)
        cargo_file.package_name = self.project_name
        cargo_file.version = "0.0.1"
        cargo_file.authors = ["you"]
        cargo_file.edition = "2018"

        for dependency in cargo_file.get_dependencies().values():
            del dependency["path"]

        cargo_file.save()

    def _patch_cargo_wasm(self):
        cargo_path = path.join(self.directory, "wasm", TemplateRust.CARGO_TOML)

        cargo_file = CargoFile(cargo_path)
        cargo_file.package_name = f"{self.project_name}-wasm"
        cargo_file.version = "0.0.1"
        cargo_file.authors = ["you"]
        cargo_file.edition = "2018"

        for dependency in cargo_file.get_dependencies().values():
            del dependency["path"]
        cargo_file.get_dependency(self.template_name)["path"] = ".."

        cargo_file.save()

        self._replace_in_files(
            [cargo_path],
            [
                (f"[dependencies.{self.template_name}]", f"[dependencies.{self.project_name}]")
            ]
        )

    def _patch_cargo_debug(self):
        cargo_debug_path = path.join(self.directory, "debug", TemplateRust.CARGO_TOML)
        if not path.exists(cargo_debug_path):
            return

        cargo_file_debug = CargoFile(cargo_debug_path)
        cargo_file_debug.package_name = f"{self.project_name}-debug"
        cargo_file_debug.version = "0.0.1"
        cargo_file_debug.authors = ["you"]
        cargo_file_debug.edition = "2018"

        for dependency in cargo_file_debug.get_dependencies().values():
            del dependency["path"]
        cargo_file_debug.get_dependency(self.template_name)["path"] = ".."

        cargo_file_debug.save()

        self._replace_in_files(
            [cargo_debug_path],
            [
                (f"[dependencies.{self.template_name}]", f"[dependencies.{self.project_name}]")
            ]
        )

    def _patch_source_code_wasm(self):
        lib_path = path.join(self.directory, "wasm", "src", "lib.rs")

        self._replace_in_files(
            [lib_path],
            [
                (f"pub use {self.template_name.replace('-', '_')}::*", f"use {self.project_name.replace('-', '_')}::*")
            ]
        )

    def _patch_source_code_debug(self):
        debug_main_path = path.join(self.directory, "debug", "src", "main.rs")
        if not path.exists(debug_main_path):
            return

        self._replace_in_files(
            [debug_main_path],
            [
                # Example: replace "use simple-erc20::*" to "use my_token::*"
                (f"use {self.template_name.replace('-', '_')}::*", f"use {self.project_name.replace('-', '_')}::*")
            ]
        )

    def _patch_source_code_tests(self):
        test_dir_path = path.join(self.directory, "tests")
        if not path.isdir(test_dir_path):
            return

        test_paths = utils.list_files(test_dir_path)
        self._replace_in_files(
            test_paths,
            [
                # Example: replace "use simple-erc20::*" to "use my_token::*"
                (f"use {self.template_name.replace('-', '_')}::*", f"use {self.project_name.replace('-', '_')}::*"),
                # Example: replace "extern crate adder;" to "extern crate myadder"
                (f"extern crate {self.template_name.replace('-', '_')};", f"extern create {self.project_name.replace('-', '_')};")
            ]
        )

    def _patch_mandos_tests(self):
        test_dir_path = path.join(self.directory, "mandos")
        if not path.isdir(test_dir_path):
            return

        test_paths = [e for e in utils.list_files(test_dir_path, suffix=".json")]
        self._replace_in_files(
            test_paths,
            [
                (f"{self.template_name}.wasm", f"{self.project_name}.wasm")
            ]
        )

        for file in test_paths:
            data = utils.read_json_file(file)
            # Patch fields
            data["name"] = data.get("name", "").replace(self.template_name, self.project_name)
            utils.write_json_file(file, data)

    def _replace_in_files(self, files, replacements):
        for file in files:
            content = utils.read_file(file)

            for to_replace, replacement in replacements:
                content = content.replace(to_replace, replacement)

            utils.write_file(file, content)


class TemplateSol(Template):
    pass



import base64
import json
import logging
from collections import OrderedDict
from typing import Any, Dict

from erdpy import errors, utils
from erdpy.accounts import Account, Address
from erdpy.interfaces import IElrondProxy, ITransaction
from erdpy.wallet import signing

logger = logging.getLogger("transactions")


class Transaction(ITransaction):
    def __init__(self):
        self.hash = ""
        self.nonce = 0
        self.value = "0"
        self.sender = ""
        self.receiver = ""
        self.gasPrice = 0
        self.gasLimit = 0
        self.data = ""
        self.chainID = ""
        self.version = 0
        self.signature = ""

    # The data field is base64-encoded. erdpy only supports utf-8 "data" at this moment.
    def data_encoded(self) -> str:
        data_bytes = self.data.encode("utf-8")
        data_base64 = base64.b64encode(data_bytes).decode()
        return data_base64

    # Useful when loading a tx from a file (when data is already encoded in base64)
    def data_decoded(self) -> str:
        return base64.b64decode(self.data).decode()

    def sign(self, account: Account):
        self.signature = signing.sign_transaction(self, account)

    def serialize(self) -> bytes:
        dictionary = self.to_dictionary()
        serialized = self._dict_to_json(dictionary)
        return serialized

    def _dict_to_json(self, dictionary: Dict[str, Any]):
        serialized = json.dumps(dictionary, separators=(',', ':')).encode("utf8")
        return serialized

    def serialize_as_inner(self):
        inner_dictionary = self.to_dictionary_as_inner()
        serialized = self._dict_to_json(inner_dictionary)
        serialized_hex = serialized.hex()
        return f"relayedTx@{serialized_hex}"

    @classmethod
    def load_from_file(cls, f: Any):
        data_json: bytes = utils.read_file(f).encode()
        fields = json.loads(data_json).get("tx")
        instance = cls()
        instance.__dict__.update(fields)
        instance.data = instance.data_decoded()
        return instance

    def dump_to(self, f: Any, extra: Any = {}):
        dump: Any = utils.Object()
        dump.tx = self.to_dictionary()
        dump.hash = self.hash or ""
        dump.data = self.data
        dump.__dict__.update(extra)
        f.writelines([dump.to_json(), "\n"])

    def send(self, proxy: IElrondProxy):
        if not self.signature:
            raise errors.TransactionIsNotSigned()

        logger.info(f"Transaction.send: nonce={self.nonce}")

        dictionary = self.to_dictionary()
        self.hash = proxy.send_transaction(dictionary)
        logger.info(f"Hash: {self.hash}")
        return self.hash

    def simulate(self, proxy: IElrondProxy):
        if not self.signature:
            raise errors.TransactionIsNotSigned()

        dictionary = self.to_dictionary()
        return proxy.simulate_transaction(dictionary)

    def to_dictionary(self) -> Dict[str, Any]:
        dictionary: Dict[str, Any] = OrderedDict()
        dictionary["nonce"] = self.nonce
        dictionary["value"] = self.value
        dictionary["receiver"] = self.receiver
        dictionary["sender"] = self.sender
        dictionary["gasPrice"] = self.gasPrice
        dictionary["gasLimit"] = self.gasLimit

        if self.data:
            dictionary["data"] = self.data_encoded()

        dictionary["chainID"] = self.chainID
        dictionary["version"] = int(self.version)

        if self.signature:
            dictionary["signature"] = self.signature

        return dictionary

    # Creates the payload for a "user" / "inner" transaction
    def to_dictionary_as_inner(self) -> Dict[str, Any]:
        dictionary = self.to_dictionary()
        dictionary["receiver"] = base64.b64encode(Address(self.receiver).pubkey()).decode()
        dictionary["sender"] = base64.b64encode(Address(self.sender).pubkey()).decode()
        dictionary["chainID"] = base64.b64encode(self.chainID.encode()).decode()
        dictionary["signature"] = base64.b64encode(bytes(bytearray.fromhex(self.signature))).decode()
        dictionary["value"] = int(self.value)

        return dictionary

    def wrap_inner(self, inner: ITransaction) -> None:
        self.data = inner.serialize_as_inner()


class BunchOfTransactions:
    def __init__(self):
        self.transactions = []

    def add_prepared(self, transaction: Transaction):
        self.transactions.append(transaction)

    def add(self, sender: Account, receiver_address: str, nonce: Any, value: Any, data: str, gas_price: int, gas_limit: int, chain: str, version: int):
        tx = Transaction()
        tx.nonce = int(nonce)
        tx.value = str(value)
        tx.sender = sender.address.bech32()
        tx.receiver = receiver_address
        tx.gasPrice = gas_price
        tx.gasLimit = gas_limit
        tx.data = data
        tx.chainID = chain
        tx.version = version

        tx.sign(sender)
        self.transactions.append(tx)

    def add_in_sequence(self):
        pass

    def send(self, proxy: IElrondProxy):
        logger.info(f"BunchOfTransactions.send: {len(self.transactions)} transactions")
        payload = [transaction.to_dictionary() for transaction in self.transactions]
        num_sent, hashes = proxy.send_transactions(payload)
        logger.info(f"Sent: {num_sent}")
        logger.info(f"TxsHashes: {hashes}")
        return num_sent, hashes


def do_prepare_transaction(args: Any) -> Transaction:
    account = Account()
    if args.pem:
        account = Account(pem_file=args.pem, pem_index=args.pem_index)
    elif args.keyfile and args.passfile:
        account = Account(key_file=args.keyfile, pass_file=args.passfile)

    tx = Transaction()
    tx.nonce = int(args.nonce)
    tx.value = args.value
    tx.sender = account.address.bech32()
    tx.receiver = args.receiver
    tx.gasPrice = int(args.gas_price)
    tx.gasLimit = int(args.gas_limit)
    tx.data = args.data
    tx.chainID = args.chain
    tx.version = int(args.version)

    tx.sign(account)
    return tx



import glob
import logging
import shutil
from os import path
from pathlib import Path
from typing import List, cast

from erdpy import dependencies, errors, myprocess, utils
from erdpy.dependencies.modules import StandaloneModule

logger = logging.getLogger("Project")


class Project:
    def __init__(self, directory):
        self.directory = str(Path(directory).resolve())

    def build(self, options=None):
        self.options = options or dict()
        self.debug = self.options.get("debug", False)
        self._ensure_dependencies_installed()
        self.perform_build()
        self._copy_build_artifacts_to_output()

    def clean(self):
        utils.remove_folder(self._get_output_folder())

    def _ensure_dependencies_installed(self):
        module_keys = self.get_dependencies()
        for module_key in module_keys:
            dependencies.install_module(module_key)

    def get_dependencies(self) -> List[str]:
        raise NotImplementedError()

    def perform_build(self) -> None:
        raise NotImplementedError()

    def get_file_wasm(self):
        return self.find_file_in_output("*.wasm")

    def find_file_globally(self, pattern):
        folder = self.directory
        return self.find_file_in_folder(folder, pattern)

    def find_file_in_output(self, pattern):
        folder = path.join(self.directory, "output")
        return self.find_file_in_folder(folder, pattern)

    def find_file_in_folder(self, folder, pattern):
        files = list(Path(folder).rglob(pattern))

        if len(files) == 0:
            raise errors.KnownError(f"No file matches pattern [{pattern}].")
        if len(files) > 1:
            logging.warning(f"More files match pattern [{pattern}]. Will pick first:\n{files}")

        file = path.join(folder, files[0])
        return Path(file).resolve()

    def _copy_build_artifacts_to_output(self) -> None:
        raise NotImplementedError()

    def _copy_to_output(self, source: str, destination: str = None):
        output_folder = self._get_output_folder()
        utils.ensure_folder(output_folder)
        destination = path.join(output_folder, destination) if destination else output_folder
        shutil.copy(source, destination)

    def _get_output_folder(self):
        return path.join(self.directory, "output")

    def get_bytecode(self):
        bytecode = utils.read_file(self.get_file_wasm(), binary=True)
        bytecode_hex = bytecode.hex()
        return bytecode_hex

    def run_tests(self, tests_directory: str, wildcard: str = ""):
        arwentools = cast(StandaloneModule, dependencies.get_module_by_key("arwentools"))
        tool_env = arwentools.get_env()
        tool = path.join(arwentools.get_parent_directory(), "mandos-test")
        test_folder = path.join(self.directory, tests_directory)

        if not wildcard:
            args = [tool, test_folder]
            myprocess.run_process(args, env=tool_env)
        else:
            pattern = path.join(test_folder, wildcard)
            test_files = glob.glob(pattern)

            for test_file in test_files:
                print("Run test for:", test_file)
                args = [tool, test_file]
                myprocess.run_process(args, env=tool_env)



from typing import Any

from erdpy.accounts import Account
from erdpy.testnet import wallets
from erdpy.testnet.config import TestnetConfiguration
from erdpy.testnet.genesis import get_delegation_address, is_foundational_node

CHAIN_ID = "local-testnet"


def build(testnet_config: TestnetConfiguration) -> Any:
    num_validators = testnet_config.num_all_validators()
    initial_nodes = []

    for nickname, [pubkey, account] in wallets.get_validators(num_validators).items():
        entry = _build_initial_nodes_entry(nickname, pubkey, account)
        initial_nodes.append(entry)

    # Then, patch the list of initial nodes, so that higher indexes will become metachain nodes.
    num_metachain_nodes = testnet_config.num_validators_in_metashard()
    num_nodes = len(initial_nodes)
    initial_nodes = initial_nodes[num_nodes - num_metachain_nodes:] + initial_nodes[:num_nodes - num_metachain_nodes]

    return {
        "startTime": testnet_config.genesis_time(),
        "roundDuration": 6000,
        "consensusGroupSize": testnet_config.shards["consensus_size"],
        "minNodesPerShard": testnet_config.shards["consensus_size"],
        "metaChainConsensusGroupSize": testnet_config.metashard["consensus_size"],
        "metaChainMinNodes": testnet_config.metashard["validators"],
        "hysteresis": 0,
        "adaptivity": False,
        "chainID": CHAIN_ID,
        "minTransactionVersion": 1,
        "initialNodes": initial_nodes
    }


def _build_initial_nodes_entry(nickname: str, pubkey: str, account: Account) -> Any:
    address = get_delegation_address().bech32() if is_foundational_node(nickname) else account.address.bech32()

    return {
        "nickname": nickname,
        "address": address,
        "pubkey": pubkey,
    }



import logging
from typing import Any

from erdpy import cli_shared, testnet

logger = logging.getLogger("cli.testnet")


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(
        subparsers,
        "testnet",
        "Set up, start and control local testnets"
    )
    subparsers = parser.add_subparsers()

    help_config_file = "An optional configuration file describing the testnet"

    # Prerequisites
    sub = cli_shared.add_command_subparser(
        subparsers,
        "testnet",
        "prerequisites",
        "Download and verify the prerequisites for running a testnet"
    )
    sub.add_argument("--configfile", type=str, required=False, default=None,
                     help=help_config_file)
    sub.set_defaults(func=testnet_prerequisites)

    # Start
    sub = cli_shared.add_command_subparser(
        subparsers,
        "testnet",
        "start",
        "Start a testnet"
    )
    sub.add_argument("--configfile", type=str, required=False, default=None,
                     help=help_config_file)
    sub.set_defaults(func=testnet_start)

    # Config
    sub = cli_shared.add_command_subparser(
        subparsers,
        "testnet",
        "config",
        "Configure a testnet (required before starting it the first time or after clean)"
    )
    sub.add_argument("--configfile", type=str, required=False, default=None,
                     help=help_config_file)
    sub.set_defaults(func=testnet_config)

    # Clean
    sub = cli_shared.add_command_subparser(
        subparsers,
        "testnet",
        "clean",
        "Erase the currently configured testnet (must be already stopped)"
    )
    sub.add_argument("--configfile", type=str, required=False, default=None,
                     help=help_config_file)
    sub.set_defaults(func=testnet_clean)


def testnet_start(args):
    logger.info("Starting testnet...")
    testnet.start(args)


def testnet_config(args):
    logger.info("Configuring testnet...")
    testnet.configure(args)


def testnet_clean(args):
    logger.info("Cleaning testnet...")
    testnet.clean(args)


def testnet_prerequisites(args):
    logger.info("Preparing prerequisites...")
    testnet.install_dependencies()



import requests
from erdpy import errors


def do_get(url):
    try:
        response = requests.get(url)
        response.raise_for_status()
        parsed = response.json()
        return get_data(parsed, url)
    except requests.HTTPError as err:
        error_data = _extract_error_from_response(err.response)
        raise errors.ProxyRequestError(url, error_data)
    except requests.ConnectionError as err:
        raise errors.ProxyRequestError(url, err)
    except Exception as err:
        raise errors.ProxyRequestError(url, err)


def do_post(url, payload):
    try:
        response = requests.post(url, json=payload)
        response.raise_for_status()
        parsed = response.json()
        return get_data(parsed, url)
    except requests.HTTPError as err:
        error_data = _extract_error_from_response(err.response)
        raise errors.ProxyRequestError(url, error_data)
    except requests.ConnectionError as err:
        raise errors.ProxyRequestError(url, err)
    except Exception as err:
        raise errors.ProxyRequestError(url, err)


def get_data(parsed, url):
    err = parsed.get("error")
    code = parsed.get("code")

    if not err and code == "successful":
        return parsed.get("data", dict())

    raise errors.ProxyRequestError(url, f"code:{code}, error: {err}")


def _extract_error_from_response(response):
    try:
        return response.json()
    except Exception:
        return response.text



from os import path

from erdpy import workstation, errors, utils


class Sink:
    def __init__(self, name, channels=None):
        self.name = name
        self.channels = channels or []

    def write(self, text):
        for channel in self.channels:
            channel.write(text)


class OutputChannel:
    def write(self, text):
        pass

    def close(self):
        pass


class ConsoleChannel(OutputChannel):
    def write(self, text):
        print(text)


class FileChannel(OutputChannel):
    def __init__(self, name):
        directory = path.join(workstation.get_tools_folder(), "OutputChannels")
        utils.ensure_folder(directory)
        filepath = path.join(directory, f"{name}.txt")
        self.file = open(filepath, "a")

    def write(self, text):
        self.file.write(text)
        self.file.write("\n")

    def close(self):
        self.file.close()


__sinks = {
    "llvm": Sink("llvm", channels=[FileChannel("llvm")])
}


def get_sink(name):
    sink = __sinks.get(name)
    if sink is None:
        raise errors.BadSink(name)
    return sink



from typing import Any, Dict


class NetworkConfig:
    def __init__(self, data: Dict[str, Any]) -> None:
        self.num_shards = data.get("erd_num_shards_without_meta", 0)
        self.min_gas_price = data.get("erd_min_gas_price", 0)
        self.chain_id = data.get("erd_chain_id", 0)
        self.min_tx_version = data.get("erd_min_transaction_version", 0)



import base64
import logging
from os import confstr

from erdpy import config, scope
from erdpy.projects import load_project
from erdpy.proxy.http_facade import do_post
from erdpy.proxy.tx_types import TxTypes

logger = logging.getLogger("proxy")


class TransactionCostEstimator:
    # needs these constant because when the observer node receive a post request from proxy with a transaction needs
    # to can figure out what type of transaction was send to can calculate an estimation
    _SENDER_ADDRESS = "erd1l453hd0gt5gzdp7czpuall8ggt2dcv5zwmfdf3sd3lguxseux2fsmsgldz"
    _RECEIVER_ADDRESS = "erd1cux02zersde0l7hhklzhywcxk4u9n4py5tdxyx7vrvhnza2r4gmq4vw35r"

    def __init__(self, proxy_url):
        self.proxy_url = proxy_url

    def estimate_tx_cost(self, arguments, tx_type):
        if tx_type == TxTypes.MOVE_BALANCE:
            return self._estimate_move_balance(arguments.data)
        elif tx_type == TxTypes.SC_DEPLOY:
            return self._estimate_sc_deploy(arguments.project)
        else:
            return self._estimate_sc_call(arguments.contract, arguments.function, arguments.arguments)

    def _estimate_move_balance(self, data):
        sender = self._SENDER_ADDRESS
        receiver = self._RECEIVER_ADDRESS
        data = data or ""
        data_bytes = base64.b64encode(data.encode())

        estimate = self._send_transaction(sender, receiver, data_bytes.decode())
        return estimate

    def _estimate_sc_deploy(self, contract_path):
        if contract_path is None:
            logger.fatal("contract-path argument missing")
            return

        project = load_project(contract_path)
        bytecode = project.get_bytecode()
        base64_bytecode = base64.b64encode(bytecode.encode())

        sender = self._SENDER_ADDRESS
        receiver = self._RECEIVER_ADDRESS
        estimate = self._send_transaction(sender, receiver, base64_bytecode.decode())
        return estimate

    def _estimate_sc_call(self, sc_address, function, arguments):
        if function is None:
            logger.fatal("function argument missing")
            return

        if sc_address is None:
            logger.fatal("sc-address argument missing")
            return

        sender = self._SENDER_ADDRESS
        receiver = sc_address
        arguments = arguments or []
        tx_data = function
        for arg in arguments:
            # TODO: call prepare_argument()
            tx_data += f"@0x0000"

        base64_bytes = base64.b64encode(tx_data.encode())
        estimate = self._send_transaction(sender, receiver, base64_bytes.decode())
        return estimate

    def _send_transaction(self, sender, receiver, data):
        tx_object = {
            "nonce": 1,
            "value": "0",
            "receiver": receiver,
            "sender": sender,
            "data": data,
            "chainID": scope.get_chain_id(),
            "version": scope.get_tx_version()
        }

        url = f"{self.proxy_url}/transaction/cost"

        raw_response = do_post(url, tx_object)
        return raw_response.get("txGasUnits", raw_response)



from erdpy import utils
from erdpy.blockatlas.core import BlockAtlas
from erdpy import cli_shared
import logging
from typing import Any


logger = logging.getLogger("cli.blockatlas")


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "blockatlas", "Interact with an Block Atlas instance")
    subparsers = parser.add_subparsers()

    parser.add_argument("--url", required=True, help="🖧 URL of Block Atlas instance")
    parser.add_argument("--coin", required=True, help="coin identifier (e.g. erd, bnb)")

    sub = cli_shared.add_command_subparser(subparsers, "blockatlas", "current-block-number", "Get latest notarized metablock (number / nonce)")
    sub.set_defaults(func=blockatlas_get_current_block_number)

    sub = cli_shared.add_command_subparser(subparsers, "blockatlas", "block-by-number", "Get block by number (nonce)")
    sub.add_argument("--number", required=True, help="the number (nonce)")
    sub.set_defaults(func=blockatlas_get_block_by_number)

    sub = cli_shared.add_command_subparser(subparsers, "blockatlas", "transactions", "Get transactions by address")
    sub.add_argument("--address", required=True, help="🖄 the address to query")
    cli_shared.add_outfile_arg(sub)
    sub.set_defaults(func=blockatlas_get_txs_by_address)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def blockatlas_get_current_block_number(args: Any) -> Any:
    client = BlockAtlas(args.url, args.coin)
    number = client.get_current_block_number()
    print(number)
    return number


def blockatlas_get_block_by_number(args: Any) -> Any:
    client = BlockAtlas(args.url, args.coin)
    block = client.get_block_by_number(args.number)
    print(block)
    return block


def blockatlas_get_txs_by_address(args: Any) -> Any:
    client = BlockAtlas(args.url, args.coin)
    transactions = client.get_txs_by_address(args.address)
    utils.dump_out_json(transactions, args.outfile)
    return transactions



from erdpy.proxy.core import ElrondProxy
from erdpy.proxy.cost import TransactionCostEstimator
import logging
from typing import Any

from erdpy import cli_contracts, cli_shared, proxy

logger = logging.getLogger("cli.cost")


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "cost", "Estimate cost of Transactions")
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(subparsers, "cost", "gas-price", "Query minimum gas price")
    cli_shared.add_proxy_arg(sub)
    sub.set_defaults(func=get_gas_price)

    sub = cli_shared.add_command_subparser(subparsers, "cost", "tx-transfer", "Query cost of regular transaction (transfer)")
    cli_shared.add_proxy_arg(sub)
    sub.add_argument("--data", required=True, help="a transaction payload, required to estimate the cost")
    sub.set_defaults(func=lambda args: get_transaction_cost(args, proxy.TxTypes.MOVE_BALANCE))

    sub = cli_shared.add_command_subparser(subparsers, "cost", "sc-deploy", "Query cost of Smart Contract deploy transaction")
    cli_shared.add_proxy_arg(sub)
    cli_contracts._add_project_or_bytecode_arg(sub)
    cli_contracts._add_arguments_arg(sub)
    sub.set_defaults(func=lambda args: get_transaction_cost(args, proxy.TxTypes.SC_DEPLOY))

    sub = cli_shared.add_command_subparser(subparsers, "cost", "sc-call", "Query cost of Smart Contract call transaction")
    cli_shared.add_proxy_arg(sub)
    cli_contracts._add_contract_arg(sub)
    cli_contracts._add_function_arg(sub)
    cli_contracts._add_arguments_arg(sub)
    sub.set_defaults(func=lambda args: get_transaction_cost(args, proxy.TxTypes.SC_CALL))

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def get_gas_price(args: Any) -> Any:
    proxy_url = args.proxy
    proxy = ElrondProxy(proxy_url)
    price = proxy.get_gas_price()
    print(price)
    return price


def get_transaction_cost(args: Any, tx_type: Any) -> Any:
    logger.debug("call_get_transaction_cost")

    cost_estimator = TransactionCostEstimator(args.proxy)
    result = cost_estimator.estimate_tx_cost(args, tx_type)
    print("Note: gas estimator is deprecated, will be updated on a future release.")
    print(result)
    return result



import base64
import logging
from typing import Any, List

from Cryptodome.Hash import keccak

from erdpy import config, errors, utils
from erdpy.accounts import Address
from erdpy.transactions import Transaction

logger = logging.getLogger("contracts")

VM_TYPE_ARWEN = "0500"


class SmartContract:
    def __init__(self, address=None, bytecode=None, metadata=None):
        self.address = Address(address)
        self.bytecode = bytecode
        self.metadata = metadata or CodeMetadata()

    def deploy(self, owner, arguments, gas_price, gas_limit, value, chain, version) -> Transaction:
        self.owner = owner
        self.compute_address()

        arguments = arguments or []
        gas_price = int(gas_price)
        gas_limit = int(gas_limit)
        value = str(value or "0")

        tx = Transaction()
        tx.nonce = owner.nonce
        tx.value = value
        tx.sender = owner.address.bech32()
        tx.receiver = Address.zero().bech32()
        tx.gasPrice = gas_price
        tx.gasLimit = gas_limit
        tx.data = self.prepare_deploy_transaction_data(arguments)
        tx.chainID = chain
        tx.version = version

        tx.sign(owner)
        return tx

    def prepare_deploy_transaction_data(self, arguments):
        tx_data = f"{self.bytecode}@{VM_TYPE_ARWEN}@{self.metadata.to_hex()}"

        for arg in arguments:
            tx_data += f"@{_prepare_argument(arg)}"

        return tx_data

    def compute_address(self):
        """
        8 bytes of zero + 2 bytes for VM type + 20 bytes of hash(owner) + 2 bytes of shard(owner)
        """
        owner_bytes = self.owner.address.pubkey()
        nonce_bytes = self.owner.nonce.to_bytes(8, byteorder="little")
        bytes_to_hash = owner_bytes + nonce_bytes
        address = keccak.new(digest_bits=256).update(bytes_to_hash).digest()
        address = bytes([0] * 8) + bytes([5, 0]) + address[10:30] + owner_bytes[30:]
        self.address = Address(address)

    def execute(self, caller, function, arguments, gas_price, gas_limit, value, chain, version) -> Transaction:
        self.caller = caller

        arguments = arguments or []
        gas_price = int(gas_price)
        gas_limit = int(gas_limit)
        value = str(value or "0")

        tx = Transaction()
        tx.nonce = caller.nonce
        tx.value = value
        tx.sender = caller.address.bech32()
        tx.receiver = self.address.bech32()
        tx.gasPrice = gas_price
        tx.gasLimit = gas_limit
        tx.data = self.prepare_execute_transaction_data(function, arguments)
        tx.chainID = chain
        tx.version = version

        tx.sign(caller)
        return tx

    def prepare_execute_transaction_data(self, function, arguments):
        tx_data = function

        for arg in arguments:
            tx_data += f"@{_prepare_argument(arg)}"

        return tx_data

    def upgrade(self, owner, arguments, gas_price, gas_limit, value, chain, version) -> Transaction:
        self.owner = owner

        arguments = arguments or []
        gas_price = int(gas_price or config.DEFAULT_GAS_PRICE)
        gas_limit = int(gas_limit)
        value = str(value or "0")

        tx = Transaction()
        tx.nonce = owner.nonce
        tx.value = value
        tx.sender = owner.address.bech32()
        tx.receiver = self.address.bech32()
        tx.gasPrice = gas_price
        tx.gasLimit = gas_limit
        tx.data = self.prepare_upgrade_transaction_data(arguments)
        tx.chainID = chain
        tx.version = version

        tx.sign(owner)
        return tx

    def prepare_upgrade_transaction_data(self, arguments):
        tx_data = f"upgradeContract@{self.bytecode}@{self.metadata.to_hex()}"

        for arg in arguments:
            tx_data += f"@{_prepare_argument(arg)}"

        return tx_data

    def query(self, proxy, function, arguments) -> List[Any]:
        arguments = arguments or []
        prepared_arguments = [_prepare_argument(argument) for argument in arguments]

        payload = {
            "ScAddress": self.address.bech32(),
            "FuncName": function,
            "Args": prepared_arguments
        }

        response = proxy.query_contract(payload)
        response_data = response.get("data", {})
        return_data = response_data.get("returnData", response_data.get("ReturnData"))
        return [self._interpret_return_data(data) for data in return_data]

    def _interpret_return_data(self, data):
        if not data:
            return data

        try:
            as_bytes = base64.b64decode(data)
            as_hex = as_bytes.hex()
            as_number = int(as_hex, 16)

            result = utils.Object()
            result.base64 = data
            result.hex = as_hex
            result.number = as_number
            return result
        except Exception:
            logger.warn(f"Cannot interpret return data: {data}")
            return None


def _prepare_argument(argument):
    hex_prefix = "0X"
    as_string = str(argument).upper()

    if as_string.startswith(hex_prefix):
        return as_string[len(hex_prefix):]

    if not as_string.isnumeric():
        raise errors.UnknownArgumentFormat(as_string)

    as_number = int(as_string)
    as_hexstring = hex(as_number)[len(hex_prefix):]
    if len(as_hexstring) % 2 == 1:
        as_hexstring = "0" + as_hexstring

    return as_hexstring


class CodeMetadata:
    def __init__(self, upgradeable=True, payable=False):
        self.upgradeable = upgradeable
        self.payable = payable

    def to_hex(self):
        return ("01" if self.upgradeable else "00") + ("02" if self.payable else "00")



import binascii
import json
import logging
from os import path
from typing import Any

from erdpy import guards
from erdpy.accounts import Account, Address
from erdpy.config import (GAS_PER_DATA_BYTE, MIN_GAS_LIMIT,
                          MetaChainSystemSCsCost)
from erdpy.errors import CannotReadValidatorsData
from erdpy.wallet.pem import parse_validator_pem
from erdpy.wallet.signing import sign_message_with_bls_key

logger = logging.getLogger("validators")

_STAKE_SMART_CONTRACT_ADDRESS = "erd1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqplllst77y4l"


def estimate_system_sc_call(args, base_cost, factor=1):
    num_bytes = len(args.data)
    gas_limit = MIN_GAS_LIMIT + num_bytes * GAS_PER_DATA_BYTE
    gas_limit += factor * base_cost
    return gas_limit


def _read_json_file(file_path):
    val_file = path.expanduser(file_path)
    guards.is_file(val_file)
    with open(file_path,  "r") as json_file:
        try:
            data = json.load(json_file)
        except Exception:
            raise CannotReadValidatorsData()
        return data


def parse_args_for_stake(args: Any):
    validators_file = args.validators_file
    validators_data = _read_json_file(validators_file)

    reward_address = args.reward_address

    if args.pem:
        account = Account(pem_file=args.pem)
    elif args.keyfile and args.passfile:
        account = Account(key_file=args.keyfile, pass_file=args.passfile)

    num_of_nodes = len(validators_data.get("validators", []))
    stake_data = 'stake@' + binascii.hexlify(num_of_nodes.to_bytes(1, byteorder="little")).decode()
    for validator in validators_data.get("validators", []):
        # get validator
        validator_pem = validator.get("pemFile")
        validator_pem = path.join(path.dirname(validators_file), validator_pem)
        seed, bls_key = parse_validator_pem(validator_pem)
        signed_message = sign_message_with_bls_key(account.address.pubkey().hex(), seed.hex())
        stake_data += f"@{bls_key}@{signed_message}"

    if reward_address:
        reward_address = Address(args.reward_address)
        stake_data += '@' + reward_address.hex()

    args.receiver = _STAKE_SMART_CONTRACT_ADDRESS
    args.data = stake_data

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.STAKE, num_of_nodes)

    return args


def parse_args_for_un_stake(args: Any):
    parsed_keys, num_keys = _parse_keys(args.nodes_public_keys)
    args.data = 'unStake' + parsed_keys
    args.receiver = _STAKE_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.UNSTAKE, num_keys)

    return args


def parse_args_for_un_bond(args: Any):
    parsed_keys, num_keys = _parse_keys(args.nodes_public_keys)
    args.data = 'unBond' + parsed_keys
    args.receiver = _STAKE_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.UNBOND, num_keys)

    return args


def parse_args_for_un_jail(args: Any):
    parsed_keys, num_keys = _parse_keys(args.nodes_public_keys)
    args.data = 'unJail' + parsed_keys
    args.receiver = _STAKE_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.UNJAIL, num_keys)

    return args


def parse_args_for_changing_reward_address(args: Any):
    reward_address = Address(args.reward_address)
    args.data = 'changeRewardAddress@' + reward_address.hex()
    args.receiver = _STAKE_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.CHANGE_REWARD_ADDRESS)

    return args


def parse_args_for_claim(args: Any):
    args.data = 'claim'
    args.receiver = _STAKE_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.CLAIM)

    return args


def _parse_keys(bls_public_keys):
    keys = bls_public_keys.split(',')
    parsed_keys = ''
    for key in keys:
        parsed_keys += '@' + key
    return parsed_keys, len(keys)



import os


def is_source_clang(directory):
    return _directory_contains_file(directory, ".c")


def is_source_cpp(directory):
    return _directory_contains_file(directory, ".cpp")


def is_source_sol(directory):
    return _directory_contains_file(directory, ".sol")


def is_source_rust(directory):
    return _directory_contains_file(directory, "Cargo.toml")


def _directory_contains_file(directory, name_suffix):
    for file in os.listdir(directory):
        if file.lower().endswith(name_suffix.lower()):
            return True



import logging
from typing import Any

from erdpy import cli_shared, wallet
from erdpy.accounts import Address
from erdpy.wallet import pem

logger = logging.getLogger("cli.wallet")


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(
        subparsers,
        "wallet",
        "Derive private key from mnemonic, bech32 address helpers etc."
    )
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(
        subparsers,
        "wallet",
        "derive",
        "Derive a PEM file from a mnemonic or generate a new PEM file (for tests only!)"
    )
    sub.add_argument("pem",
                     help="path of the output PEM file")
    sub.add_argument("--mnemonic", action="store_true",
                     help="whether to derive from an existing mnemonic")
    sub.add_argument("--index",
                     help="the account index", type=int, default=0)
    sub.set_defaults(func=generate_pem)

    sub = cli_shared.add_command_subparser(
        subparsers,
        "wallet",
        "bech32",
        "Helper for encoding and decoding bech32 addresses"
    )
    sub.add_argument("value",
                     help="the value to encode or decode")
    group = sub.add_mutually_exclusive_group(required=True)
    group.add_argument("--encode", action="store_true",
                       help="whether to encode")
    group.add_argument("--decode", action="store_true",
                       help="whether to decode")
    sub.set_defaults(func=do_bech32)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def generate_pem(args: Any):
    pem_file = args.pem
    mnemonic = args.mnemonic
    index = args.index

    seed, pubkey = wallet.generate_pair()
    if mnemonic:
        mnemonic = input("Enter mnemonic:\n")
        mnemonic = mnemonic.strip()
        seed, pubkey = wallet.derive_keys(mnemonic, index)

    address = Address(pubkey)
    pem.write(pem_file, seed, pubkey, name=address.bech32())
    logger.info(f"Created PEM file [{pem_file}] for [{address.bech32()}]")


def do_bech32(args: Any):
    encode = args.encode
    value = args.value
    address = Address(value)

    result = address.bech32() if encode else address.hex()
    print(result)
    return result




from typing import Any, List

from erdpy.accounts import Account, Address
from erdpy.testnet import wallets
from erdpy.testnet.config import TestnetConfiguration
from erdpy.testnet.genesis import (get_delegation_address,
                                   is_foundational_node, is_last_user)

ENTIRE_SUPPLY = 20000000000000000000000000
# For testnet, we delegate for 1 node
DELEGATED_VALUE = 2500000000000000000000


def build(testnet_config: TestnetConfiguration) -> List[Any]:
    num_validators = testnet_config.num_all_validators()
    genesis_items = []
    remaining_supply = ENTIRE_SUPPLY
    remaining_to_delegate = DELEGATED_VALUE
    delegation_address = get_delegation_address()

    for nickname, account in wallets.get_validator_wallets(num_validators).items():
        if is_foundational_node(nickname):
            continue

        value = 2500000000000000000000
        entry = _build_validator_entry(nickname, account, value)
        genesis_items.append(entry)
        remaining_supply -= value

    for nickname, account in wallets.get_users().items():
        # The last user (mike) gets all remaining tokens
        value = remaining_supply if is_last_user(nickname) else 100000000000000000000000
        delegated_value = remaining_to_delegate if is_last_user(nickname) else 100000000000000000000
        entry = _build_user_entry(nickname, account, value, delegated_value, delegation_address)
        genesis_items.append(entry)
        remaining_supply -= value
        remaining_to_delegate -= delegated_value

    return genesis_items


def _build_validator_entry(nickname: str, account: Account, value: int):
    return {
        "nickname": nickname,
        "address": account.address.bech32(),
        "supply": str(value),
        "balance": "0",
        "stakingvalue": str(value),
        "delegation": {
            "address": "",
            "value": "0"
        }
    }


def _build_user_entry(nickname: str, account: Account, value: int, delegated_value: int, delegation_address: Address):
    return {
        "nickname": nickname,
        "address": account.address.bech32(),
        "supply": str(value),
        "balance": str(value - delegated_value),
        "stakingvalue": "0",
        "delegation": {
            "address": delegation_address.bech32(),
            "value": str(delegated_value)
        }
    }



from erdpy.transactions import do_prepare_transaction
from erdpy.proxy.core import ElrondProxy
from erdpy import validators
from typing import Any

from erdpy import cli_shared


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "validator", "Stake, Unjail and other actions useful for "
                                                                     "Validators")
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(subparsers, "validator", "stake", "Stake value into the Network")
    _add_common_arguments(sub)
    sub.add_argument("--reward-address", default="", help="the reward address")
    sub.add_argument("--validators-file", required=True, help="a JSON file describing the Nodes")
    sub.set_defaults(func=do_stake)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "unstake", "Unstake value")
    _add_common_arguments(sub)
    _add_nodes_arg(sub)
    sub.set_defaults(func=do_unstake)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "unjail", "Unjail a Validator Node")
    _add_common_arguments(sub)
    _add_nodes_arg(sub)
    sub.set_defaults(func=do_unjail)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "unbond", "Unbond")
    _add_common_arguments(sub)
    _add_nodes_arg(sub)
    sub.set_defaults(func=do_unbond)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "change-reward-address", "Change the reward address")
    _add_common_arguments(sub)
    sub.add_argument("--reward-address", required=True, help="the new reward address")
    sub.set_defaults(func=change_reward_address)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "claim", "Claim rewards")
    _add_common_arguments(sub)
    sub.set_defaults(func=do_claim)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def _add_common_arguments(sub: Any):
    cli_shared.add_proxy_arg(sub)
    cli_shared.add_wallet_args(sub)
    cli_shared.add_tx_args(sub, with_receiver=False, with_data=False, with_estimate_gas=True)
    cli_shared.add_outfile_arg(sub, what="signed transaction, hash")


def _add_nodes_arg(sub: Any):
    sub.add_argument("--nodes-public-keys", required=True, help="the public keys of the nodes as CSV (addrA,addrB)")


def do_stake(args: Any):
    cli_shared.prepare_nonce_in_args(args)
    args = validators.parse_args_for_stake(args)
    tx = do_prepare_transaction(args)

    try:
        tx.send(ElrondProxy(args.proxy))
    finally:
        tx.dump_to(args.outfile)


def do_unstake(args: Any):
    cli_shared.prepare_nonce_in_args(args)
    args = validators.parse_args_for_un_stake(args)
    tx = do_prepare_transaction(args)

    try:
        tx.send(ElrondProxy(args.proxy))
    finally:
        tx.dump_to(args.outfile)


def do_unjail(args: Any):
    cli_shared.prepare_nonce_in_args(args)
    args = validators.parse_args_for_un_jail(args)
    tx = do_prepare_transaction(args)

    try:
        tx.send(ElrondProxy(args.proxy))
    finally:
        tx.dump_to(args.outfile)


def do_unbond(args: Any):
    cli_shared.prepare_nonce_in_args(args)
    args = validators.parse_args_for_un_bond(args)
    tx = do_prepare_transaction(args)

    try:
        tx.send(ElrondProxy(args.proxy))
    finally:
        tx.dump_to(args.outfile)


def change_reward_address(args: Any):
    cli_shared.prepare_nonce_in_args(args)
    args = validators.parse_args_for_changing_reward_address(args)
    tx = do_prepare_transaction(args)

    try:
        tx.send(ElrondProxy(args.proxy))
    finally:
        tx.dump_to(args.outfile)


def do_claim(args: Any):
    cli_shared.prepare_nonce_in_args(args)
    args = validators.parse_args_for_claim(args)
    tx = do_prepare_transaction(args)

    try:
        tx.send(ElrondProxy(args.proxy))
    finally:
        tx.dump_to(args.outfile)



import logging
import os
import subprocess
from os import path
from pathlib import Path

from erdpy import dependencies, errors, myprocess, utils
from erdpy.projects.project_base import Project

logger = logging.getLogger("ProjectClang")


class ProjectClang(Project):
    def __init__(self, directory):
        super().__init__(directory)

    def perform_build(self):
        self.build_configuration = ClangBuildConfiguration(self, self.debug)
        self.unit = self.find_file_globally("*.c")
        self.file_ll = self.unit.with_suffix(".ll")
        self.file_o = self.unit.with_suffix(".o")
        self.file_export = self.unit.with_suffix(".export")

        try:
            self._do_clang()
            self._do_llc()
            self._do_wasm()
        except subprocess.CalledProcessError as err:
            raise errors.BuildError(err.output)

    def _do_clang(self):
        logger.info("_do_clang")
        tool = path.join(self._get_llvm_path(), "clang-9")
        args = [
            tool,
            "-cc1", "-emit-llvm",
            "-triple=wasm32-unknown-unknown-wasm",
        ]
        if self.options.get("optimized", False):
            args.append("-Ofast")
        else:
            args.append("-O0")
        args.append(str(self.unit))
        myprocess.run_process(args)

    def _do_llc(self):
        logger.info("_do_llc")
        tool = path.join(self._get_llvm_path(), "llc")
        args = [tool]
        if self.options.get("optimized", False):
            args.append("-O3")
        else:
            args.append("-O0")
        args.extend(["-filetype=obj", self.file_ll, "-o", self.file_o])
        myprocess.run_process(args)

    def _do_wasm(self):
        logger.info("_do_wasm")
        tool = path.join(self._get_llvm_path(), "wasm-ld")
        args = [
            tool,
            "--no-entry",
            str(self.file_o),
            "-o", self.find_file_globally("*.c").with_suffix(".wasm"),
            "--strip-all",
            "-allow-undefined"
        ]

        if self.options.get("verbose", False):
            args.append("--verbose")

        for export in self.build_configuration.exports:
            args.append(f"-export={export}")

        myprocess.run_process(args)

    def _copy_build_artifacts_to_output(self):
        source_file = self.find_file_globally("*.c")
        self._copy_to_output(source_file.with_suffix(".wasm"))
        os.remove(source_file.with_suffix(".wasm"))
        os.remove(source_file.with_suffix(".ll"))
        os.remove(source_file.with_suffix(".o"))

    def _get_llvm_path(self):
        return dependencies.get_module_directory("llvm")

    def get_dependencies(self):
        return ["llvm"]


class ClangBuildConfiguration:
    def __init__(self, project, debug):
        self.project = project
        self.debug = debug
        self.exports = self._get_exports()

    def _get_exports(self):
        file_export = self.project.find_file_globally("*.export")
        lines = utils.read_lines(file_export)
        return lines




import base64
from binascii import b2a_base64, hexlify, unhexlify
from json import load

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, hmac
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from erdpy import errors


# References:
# Thanks for this implementation @flyingbasalt
# https://github.com/flyingbasalt/erdkeys
def load_from_key_file(key_file_json, password):
    with open(key_file_json) as json_f:
        keystore = load(json_f)

    backend = default_backend()

    # derive the decryption key
    kdf_name = keystore['crypto']['kdf']
    if kdf_name != 'scrypt':
        raise errors.UnknownDerivationFunction()

    salt = unhexlify(keystore['crypto']['kdfparams']['salt'])
    dklen = keystore['crypto']['kdfparams']['dklen']
    n = keystore['crypto']['kdfparams']['n']
    p = keystore['crypto']['kdfparams']['p']
    r = keystore['crypto']['kdfparams']['r']

    kdf = Scrypt(salt=salt, length=dklen, n=n, r=r, p=p, backend=backend)
    key = kdf.derive(bytes(password.encode()))

    # decrypt the private key with half of the decryption key
    cipher_name = keystore['crypto']['cipher']
    if cipher_name != 'aes-128-ctr':
        raise errors.UnknownCipher(name=cipher_name)

    iv = unhexlify(keystore['crypto']['cipherparams']['iv'])
    ciphertext = unhexlify(keystore['crypto']['ciphertext'])
    decryption_key = key[0:16]

    cipher = Cipher(algorithms.AES(decryption_key), modes.CTR(iv), backend=backend)
    decryptor = cipher.decryptor()
    plaintext = decryptor.update(ciphertext) + decryptor.finalize()
    pemified_private_key = b2a_base64(hexlify(plaintext))

    hmac_key = key[16:32]
    h = hmac.HMAC(hmac_key, hashes.SHA256(), backend=backend)
    h.update(ciphertext)
    mac = h.finalize()

    if mac != unhexlify(keystore['crypto']['mac']):
        raise errors.InvalidKeystoreFilePassword()

    address_bech32 = keystore['bech32']
    private_key = ''.join([pemified_private_key[i:i + 64].decode() for i in range(0, len(pemified_private_key), 64)])

    key_hex = base64.b64decode(private_key).decode()
    key_bytes = bytes.fromhex(key_hex)

    seed = key_bytes[:32]

    return address_bech32, seed


def get_password(pass_file):
    with open(pass_file) as pass_f:
        return pass_f.read().strip()



import setuptools

with open("README.md", "r") as fh:
    long_description = "https://github.com/ElrondNetwork/erdpy"

# See https://packaging.python.org/tutorials/packaging-projects/
setuptools.setup(
    name="erdpy",
    version="0.9.2",
    description="Elrond Smart Contracts Tools and Python SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ElrondNetwork/erdpy",
    author="Elrond Network",
    license="GPL",
    packages=setuptools.find_packages(
        include=["erdpy*"], exclude=["examples*"]),
    include_package_data=True,
    setup_requires=["wheel"],
    install_requires=[
        "toml==0.10.0", "bottle", "requests", "pynacl", "pycryptodomex", "cryptography"
    ],
    zip_safe=False,
    keywords=["Elrond"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: POSIX :: Linux",
        "Intended Audience :: Developers",
        "Development Status :: 3 - Alpha"
    ],
    entry_points={
        "console_scripts": [
            "erdpy=erdpy.cli:main",
        ],
    },
    python_requires=">=3.6"
)



from erdpy.dispatcher.transactions.queue import TransactionQueue
import logging
from typing import Any

from erdpy import cli_shared

logger = logging.getLogger("cli.dispatcher")


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "dispatcher", "Enqueue transactions, then bulk dispatch them")
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(subparsers, "dispatcher", "enqueue", "Enqueue a transaction")
    cli_shared.add_tx_args(sub, with_nonce=False)
    sub.set_defaults(func=enqueue_transaction)

    sub = cli_shared.add_command_subparser(subparsers, "dispatcher", "dispatch", "Dispatch queued transactions")
    cli_shared.add_proxy_arg(sub)
    cli_shared.add_wallet_args(sub)
    sub.set_defaults(func=dispatch_transactions)

    sub = cli_shared.add_command_subparser(subparsers, "dispatcher", "dispatch-continuously", "Continuously dispatch queued transactions")
    cli_shared.add_proxy_arg(sub)
    cli_shared.add_wallet_args(sub)
    sub.add_argument("--interval", required=True, help="the interval to retrieve transactions from the queue, in seconds")
    sub.set_defaults(func=dispatch_transactions_continuously)

    sub = cli_shared.add_command_subparser(subparsers, "dispatcher", "clean", "Clear queue of transactions")
    sub.set_defaults(func=clean_transactions_queue)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def enqueue_transaction(args: Any):
    queue = TransactionQueue()
    queue.enqueue_transaction(args)


def dispatch_transactions(args: Any):
    queue = TransactionQueue()
    queue.dispatch_transactions(args)


def dispatch_transactions_continuously(args: Any):
    queue = TransactionQueue()
    queue.dispatch_transactions_continuously(args)


def clean_transactions_queue():
    queue = TransactionQueue()
    queue.clean_transactions_queue()



from erdpy.wallet.core import derive_keys, mnemonic_to_bip39seed, bip39seed_to_private_key
from erdpy.wallet.generator import generate_pair


__all__ = ["derive_keys", "mnemonic_to_bip39seed", "generate_pair", "bip39seed_to_private_key"]



import logging
import os
from os import path

from erdpy import config
from erdpy.testnet.config import TestnetConfiguration
from erdpy.testnet.nodes_setup_json import CHAIN_ID

chain_id = None
proxy = None

logger = logging.getLogger("scope")


def initialize():
    cwd = os.getcwd()
    testnet_toml = path.join(cwd, "testnet.toml")

    if os.path.exists(testnet_toml):
        logger.info("running in local testnet")

        global chain_id, proxy

        testnet_config = TestnetConfiguration.from_file(testnet_toml)
        chain_id = CHAIN_ID
        proxy = f"http://localhost:{testnet_config.proxy_port()}"

    logger.info(f"initialized. chain = {chain_id}, proxy = {proxy}")


def get_chain_id():
    return chain_id or config.get_chain_id()


def get_tx_version() -> int:
    return config.get_tx_version()


def get_proxy() -> str:
    return proxy or config.get_proxy()



from erdpy.validators.core import (estimate_system_sc_call,
                                   parse_args_for_changing_reward_address,
                                   parse_args_for_claim, parse_args_for_stake,
                                   parse_args_for_un_bond,
                                   parse_args_for_un_jail,
                                   parse_args_for_un_stake)

__all__ = ["estimate_system_sc_call", "parse_args_for_stake", "parse_args_for_un_stake", "parse_args_for_un_bond", "parse_args_for_un_jail", "parse_args_for_changing_reward_address", "parse_args_for_claim"]



import logging
import subprocess
from os import path
from pathlib import Path
from typing import Any, MutableMapping

from erdpy import dependencies, errors, myprocess, utils
from erdpy.projects.project_base import Project

logger = logging.getLogger("ProjectRust")


class ProjectRust(Project):
    def __init__(self, directory):
        super().__init__(directory)
        self.cargo_file = self._get_cargo_file()

    def clean(self):
        super().clean()
        utils.remove_folder(path.join(self.directory, "wasm", "target"))

    def _get_cargo_file(self):
        cargo_path = path.join(self.directory, "Cargo.toml")
        return CargoFile(cargo_path)

    def perform_build(self):
        try:
            self.run_cargo()
        except subprocess.CalledProcessError as err:
            raise errors.BuildError(err.output)

    def run_cargo(self):
        env = self._get_env()

        if self.debug:
            args = ["cargo", "build"]
        else:
            args = [
                "cargo",
                "build",
                "--target=wasm32-unknown-unknown",
                "--release"
            ]

            env["RUSTFLAGS"] = "-C link-arg=-s"

        cwd = path.join(self.directory, "wasm")
        result = myprocess.run_process_async(args, env=env, cwd=cwd)
        if result != 0:
            raise errors.BuildError(f"error code = {result}, see output")

    def _copy_build_artifacts_to_output(self):
        name = self.cargo_file.package_name.replace("-", "_")
        name_with_suffix = f"{name}_wasm.wasm"
        name_without_suffix = f"{self.cargo_file.package_name}.wasm"
        wasm_file = Path(self.directory, "wasm", "target", "wasm32-unknown-unknown", "release", name_with_suffix).resolve()
        self._copy_to_output(wasm_file, name_without_suffix)

    def get_dependencies(self):
        return ["rust"]

    def _get_env(self):
        return dependencies.get_module_by_key("rust").get_env()


class CargoFile:
    def __init__(self, path):
        self.data = {}
        self.path = path

        try:
            self._parse_file()
        except Exception as err:
            raise errors.BuildError("Can't read or parse [Cargo.toml] file", err)

    def _parse_file(self):
        self.data = utils.read_toml_file(self.path)

    @property
    def package_name(self):
        return self._get_package().get("name")

    @package_name.setter
    def package_name(self, value):
        self._get_package().update({"name": value})

    @property
    def version(self):
        return self._get_package().get("version")

    @version.setter
    def version(self, value):
        self._get_package().update({"version": value})

    @property
    def authors(self):
        return self._get_package().get("authors")

    @authors.setter
    def authors(self, value):
        self._get_package().update({"authors": value})

    @property
    def edition(self):
        return self._get_package().get("edition")

    @edition.setter
    def edition(self, value):
        self._get_package().update({"edition": value})

    def save(self):
        utils.write_toml_file(self.path, self.data)

    def _get_package(self) -> MutableMapping[str, Any]:
        if "package" not in self.data:
            self.data["package"] = {}
        return self.data["package"]

    def get_dependencies(self) -> MutableMapping[str, Any]:
        if "dependencies" not in self.data:
            self.data["dependencies"] = {}
        return self.data["dependencies"]

    def get_dependency(self, name) -> MutableMapping[str, Any]:
        dependencies = self.get_dependencies()
        dependency = dependencies.get(name)
        if dependency is None:
            raise errors.BuildError(f"Can't get cargo dependency: {name}")
        return dependency



from typing import Any

from erdpy.testnet import genesis
from erdpy.testnet.config import TestnetConfiguration


def patch(data: Any, testnet_config: TestnetConfiguration):
    owner = genesis.get_owner_of_genesis_contracts()

    data[0]["owner"] = owner.address.bech32()
    data[1]["owner"] = owner.address.bech32()



import logging
from os import path
from typing import Any

import nacl.encoding
import nacl.signing
from erdpy import dependencies, myprocess
from erdpy.errors import CannotSignMessageWithBLSKey
from erdpy.interfaces import IAccount, ITransaction

logger = logging.getLogger("wallet")


def sign_transaction(transaction: ITransaction, account: IAccount) -> str:
    seed: bytes = account.get_seed()
    signing_key: Any = nacl.signing.SigningKey(seed)

    data_json = transaction.serialize()
    signed = signing_key.sign(data_json)
    signature = signed.signature
    signature_hex = signature.hex()

    return signature_hex


def sign_message_with_bls_key(message, seed):
    dependencies.install_module("mcl_signer")
    tool = path.join(dependencies.get_module_directory("mcl_signer"), "signer")

    try:
        signed_message = myprocess.run_process([tool, message, seed], dump_to_stdout=False)
        return signed_message
    except Exception:
        raise CannotSignMessageWithBLSKey()




import logging
import sys

import requests

from erdpy import errors

logger = logging.getLogger("downloader")

CHUNK_SIZE = 1024 * 16


def download(url: str, filename: str) -> None:
    if url is None:
        raise errors.BadUrlError()

    logger.info(f"download_url.url: {url}")
    logger.info(f"download_url.filename: {filename}")

    try:
        response = requests.get(url, stream=True)
        response.raise_for_status()
        total_size = int(response.headers.get("content-length", 0))
        chunk_number = 0
        progress = 0
        with open(filename, 'wb') as file:
            for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                file.write(chunk)
                progress = _report_download_progress(progress, chunk_number, total_size)
                chunk_number += 1
            print("100 %", end="\r", file=sys.stderr)
            print()
    except requests.HTTPError as err:
        raise errors.DownloadError(
            f"Could not download [{url}] to [{filename}]") from err

    logger.info("Download done.")


def _report_download_progress(progress, chunk_number, total_size):
    num_chunks = total_size / CHUNK_SIZE + 1
    new_progress = int(chunk_number / num_chunks * 100)
    if new_progress % 10 == 0:
        print(f"{progress} %", end="\r", file=sys.stderr)
    return new_progress
